
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <math.h>

#include <png.h>

#include "PNG.hh"

static void
readData( png_structp ctx, png_bytep buffer, png_size_t size )
{
    std::ifstream* src = (std::ifstream*) ::png_get_io_ptr( ctx );
    src->read( (char*)buffer, size );
}

static void
writeData( png_structp ctx, png_bytep buffer, png_size_t size )
{
    std::ofstream* dst = (std::ofstream*) ::png_get_io_ptr( ctx );
    dst->write( (char*)buffer, size );
}

static unsigned int
clip( double value, unsigned int max )
{
    int cc = (int)( value + 0.5 );

    if ( cc < 0 ) {
	cc = 0;
    } else if ( (unsigned int)cc > max ) {
	cc = (int)max;
    }

    return (unsigned int)cc;
}

static void
flushData( png_structp ctx )
{
    std::ofstream* dst = (std::ofstream*) ::png_get_io_ptr( ctx );
    *dst << std::flush;
}

PNG::PNG( void ) : pixels( 0 )
{
}

PNG::~PNG( void )
{
    delete[] this->pixels;
}

PNG::PNG( unsigned int ww, unsigned int hh, unsigned int ch )
{
    this->width = ww;
    this->height = hh;
    this->channels = ch;

    this->pixels = new unsigned char[
	    this->width * this->height * this->channels
	];
    ::memset( this->pixels, 0, this->width * this->height * this->channels );
}

PNG::PNG( const PNG& other ) : pixels( 0 )
{
    *this = other;
}

PNG&
PNG::operator = ( const PNG& other )
{
    delete[] this->pixels;

    this->width = other.width;
    this->height = other.height;
    this->channels = other.channels;

    this->pixels = new unsigned char[
	    this->width * this->height * this->channels
	];
    if ( this->pixels != 0 ) {
	::memcpy(
		this->pixels, other.pixels,
		this->width * this->height * this->channels
	    );
    }

    return *this;
}

bool
PNG::read( const char* fileName )
{
    std::ifstream src( fileName );

    if ( src.bad() || src.fail() ) {
	return false;
    }

    png_structp pngPtr = ::png_create_read_struct(
	    PNG_LIBPNG_VER_STRING, 0, 0, 0
	);

    if ( pngPtr == 0 ) {
	std::cerr << "Unable to create png read structure" << std::endl;
	return false;
    }

    png_infop infoPtr = ::png_create_info_struct( pngPtr );
    if ( infoPtr == 0 ) {
	std::cerr << "Unable to create png info struct" << std::endl;
	::png_destroy_read_struct( &pngPtr, 0, 0 );
	return false;
    }

    if ( ::setjmp( pngPtr->jmpbuf ) ) {
	std::cerr << "Something wrong during read png" << std::endl;
	::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );
	return false;
    }

    ::png_set_read_fn( pngPtr, &src, ::readData );
    ::png_read_info( pngPtr, infoPtr );

    png_uint_32 _width = 0;
    png_uint_32 _height = 0;
    int bpp = 0;
    int colorType = 0;
    int interlaceType = 0;

    ::png_get_IHDR( pngPtr, infoPtr,
	&_width, &_height, &bpp, &colorType, &interlaceType, 0, 0
    );

    this->width = _width;
    this->height = _height;

    if ( colorType == PNG_COLOR_TYPE_PALETTE && bpp <= 8 ) {
	::png_set_expand( pngPtr );
    }
    if ( colorType == PNG_COLOR_TYPE_GRAY and bpp < 8 ) {
	::png_set_expand( pngPtr );
    }
    if ( ::png_get_valid( pngPtr, infoPtr, PNG_INFO_tRNS ) ) {
	::png_set_expand( pngPtr );
    }
    if ( ( colorType & PNG_COLOR_MASK_ALPHA ) != 0 ) {
	::png_set_strip_alpha( pngPtr );
    }

    if ( colorType == PNG_COLOR_TYPE_GRAY
    or colorType == PNG_COLOR_TYPE_GRAY_ALPHA ) {
	::png_set_gray_to_rgb( pngPtr );
    }

    if ( bpp == 16 ) {
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	    ::png_set_swap( pngPtr );
#endif
	::png_set_strip_16( pngPtr );
    }

    unsigned int nPasses = ::png_set_interlace_handling( pngPtr );

    ::png_read_update_info( pngPtr, infoPtr );

    this->channels = infoPtr->channels;

    this->pixels = new unsigned char[
	    this->width * this->height * this->channels
	];

    if ( this->pixels == 0 ) {
	::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );
	return false;
    }

    png_bytep* rowPtrs = new png_bytep[ this->height ];

    if ( rowPtrs == 0 ) {
	delete[] this->pixels;
	this->pixels = 0;
	::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );
	return false;
    }


    unsigned char* ptr = &this->pixels[ 0 ];

    for ( unsigned int ii=0; ii < height; ++ii ) {
	rowPtrs[ii] = (png_bytep)ptr;
	ptr = &ptr[ this->width * this->channels ];
    }

    ::png_read_image( pngPtr, rowPtrs );

    delete[] rowPtrs;

    ::png_read_end( pngPtr, infoPtr );
    ::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );

    return true;
}

bool
PNG::readAlpha( const char* fileName )
{
    std::ifstream src( fileName );

    if ( src.bad() || src.fail() ) {
	return false;
    }

    png_structp pngPtr = ::png_create_read_struct(
	    PNG_LIBPNG_VER_STRING, 0, 0, 0
	);

    if ( pngPtr == 0 ) {
	std::cerr << "Unable to create png read structure" << std::endl;
	return false;
    }

    png_infop infoPtr = ::png_create_info_struct( pngPtr );
    if ( infoPtr == 0 ) {
	std::cerr << "Unable to create png info struct" << std::endl;
	::png_destroy_read_struct( &pngPtr, 0, 0 );
	return false;
    }

    if ( ::setjmp( pngPtr->jmpbuf ) ) {
	std::cerr << "Something wrong during read png" << std::endl;
	::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );
	return false;
    }

    ::png_set_read_fn( pngPtr, &src, ::readData );
    ::png_read_info( pngPtr, infoPtr );

    png_uint_32 _width = 0;
    png_uint_32 _height = 0;
    int bpp = 0;
    int colorType = 0;
    int interlaceType = 0;

    ::png_get_IHDR( pngPtr, infoPtr,
	&_width, &_height, &bpp, &colorType, &interlaceType, 0, 0
    );

    this->width = _width;
    this->height = _height;

    if ( colorType == PNG_COLOR_TYPE_PALETTE && bpp <= 8 ) {
	::png_set_expand( pngPtr );
    }
    if ( colorType == PNG_COLOR_TYPE_GRAY and bpp < 8 ) {
	::png_set_expand( pngPtr );
    }
    if ( ::png_get_valid( pngPtr, infoPtr, PNG_INFO_tRNS ) ) {
	::png_set_expand( pngPtr );
    }

    if ( colorType == PNG_COLOR_TYPE_GRAY
    or colorType == PNG_COLOR_TYPE_GRAY_ALPHA ) {
	::png_set_gray_to_rgb( pngPtr );
    }

    if ( bpp == 16 ) {
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	    ::png_set_swap( pngPtr );
#endif
	::png_set_strip_16( pngPtr );
    }

    unsigned int nPasses = ::png_set_interlace_handling( pngPtr );

    ::png_read_update_info( pngPtr, infoPtr );

    this->channels = infoPtr->channels;

    this->pixels = new unsigned char[
	    this->width * this->height * this->channels
	];

    if ( this->pixels == 0 ) {
	::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );
	return false;
    }

    png_bytep* rowPtrs = new png_bytep[ this->height ];

    if ( rowPtrs == 0 ) {
	delete[] this->pixels;
	this->pixels = 0;
	::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );
	return false;
    }


    unsigned char* ptr = &this->pixels[ 0 ];

    for ( unsigned int ii=0; ii < height; ++ii ) {
	rowPtrs[ii] = (png_bytep)ptr;
	ptr = &ptr[ this->width * this->channels ];
    }

    ::png_read_image( pngPtr, rowPtrs );

    delete[] rowPtrs;

    ::png_read_end( pngPtr, infoPtr );
    ::png_destroy_read_struct( &pngPtr, &infoPtr, 0 );

    return true;
}

bool
PNG::write( const char* fileName )
{
    std::ofstream dst( fileName );

    if ( dst.bad() || dst.fail() ) {
	return false;
    }

    png_structp pngPtr = ::png_create_write_struct(
	    PNG_LIBPNG_VER_STRING, 0, 0, 0
	);

    if ( pngPtr == 0 ) {
	std::cerr << "Unable to create png write structure" << std::endl;
	return false;
    }

    png_infop infoPtr = ::png_create_info_struct( pngPtr );
    if ( infoPtr == 0 ) {
	std::cerr << "Unable to create png info struct" << std::endl;
	::png_destroy_write_struct( &pngPtr, 0 );
	return false;
    }

    if ( ::setjmp( pngPtr->jmpbuf ) ) {
	std::cerr << "Something wrong during write png" << std::endl;
	::png_destroy_write_struct( &pngPtr, &infoPtr );
	return false;
    }

    ::png_set_write_fn( pngPtr, &dst, ::writeData, ::flushData );

    ::png_set_compression_level( pngPtr, Z_BEST_COMPRESSION );
    // ::png_set_compression_method( pngPtr, 8 );

    ::png_set_IHDR( pngPtr, infoPtr, this->width, this->height, 8,
	    PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
	    PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT
	);

    ::png_write_info( pngPtr, infoPtr );

    for ( unsigned int ii=0; ii < this->height; ++ii ) {
	png_bytep rowPtr = (png_bytep)&this->pixels[
		ii * this->width * this->channels
	    ];
	::png_write_row( pngPtr, rowPtr );
    }

    ::png_write_end( pngPtr, infoPtr );

    ::png_destroy_write_struct( &pngPtr, &infoPtr );

    return true;
}
