
#include <iostream>
#include <string>
#include <cmath>
#include <cstdlib>

#include "vec.hh"
#include "ray.hh"
#include "mirror.hh"
#include "scope.hh"
#include "PNG.hh"
#include "seq.hh"
#include "progress.hh"

static unsigned int clip( double _dd, unsigned int _max = 255 ) {
    int vv = (int)( _dd * (double)_max + 0.5 );
    if ( vv < 0 ) {
	return 0;
    } else if ( vv > _max ) {
	return (unsigned int)_max;
    } else {
	return vv;
    }
}

int
main( int _argc, const char* const _argv[] ) {
    unsigned int ww = 0;
    unsigned int hh = 0;
    double cc = 1.0;
    double length = 5.0;
    double fps = 29.97;

    const char* iprefix = "./in/base ";
    unsigned int ifill = 3;
    const char* isuffix = ".png";

    const char* oprefix = "./out/base ";
    unsigned int ofill = 3;
    const char* osuffix = ".png";

	/*
	** parse arguments
	*/
    for ( int ii=1; ii < _argc; ++ii ) {
	std::string arg( _argv[ ii ] );

	if ( arg == "--size" || arg == "-s" ) {
	    ww = std::strtol( _argv[ ++ii ], 0, 0 );
	    hh = std::strtol( _argv[ ++ii ], 0, 0 );
	    if ( ww < 1 || hh < 1 ) {
		std::cerr << "Width and height must both be positive integers";
		std::cerr << std::endl;
		return __LINE__;
	    }
	} else if ( arg == "--length" || arg == "-l" ) {
	    length = std::strtod( _argv[ ++ii ], 0 );
	    if ( length < 0.0 ) {
		std::cerr << "Length must be non-negative";
		std::cerr << std::endl;
		return __LINE__;
	    }
	} else if ( arg == "--c" || arg == "-c" ) {
	    cc = std::strtod( _argv[ ++ii ], 0 );
	    if ( cc <= 0.0 ) {
		std::cerr << "Lightspeed must be positive";
		std::cerr << std::endl;
		return __LINE__;
	    }
	} else if ( arg == "--informat" || arg == "-i" ) {
	    iprefix = _argv[ ++ii ];
	    ifill = std::strtol( _argv[ ++ii ], 0, 0 );
	    isuffix = _argv[ ++ii ];
	} else if ( arg == "--outformat" || arg == "-o" ) {
	    oprefix = _argv[ ++ii ];
	    ofill = std::strtol( _argv[ ++ii ], 0, 0 );
	    osuffix = _argv[ ++ii ];
	} else if ( arg == "--fps" ) {
	    fps = std::strtod( _argv[ ++ii ], 0 );
	    if ( fps <= 0.0 ) {
		std::cerr << "Frames-per-second must be positive...";
		std::cerr << std::endl;
		return __LINE__;
	    }
	} else {
	    if ( arg != "--help" && arg != "-h" && arg != "-?" ) {
		std::cerr << "Unrecogized arg: " << arg << std::endl;
	    }
	    std::cerr << "Usage: " << _argv[0] << std::endl;
	    std::cerr << "        [--informat pre fill post]" << std::endl;
	    std::cerr << "        [--outformat pre fill post]" << std::endl;
	    std::cerr << "        [--size ww hh]" << std::endl;
	    std::cerr << "        [--length ll.lll]" << std::endl;
	    std::cerr << "        [--c ss.sss]" << std::endl;
	    std::cerr << "        [--fps ff.fff]" << std::endl;
	    return __LINE__;
	}
    }

	/*
	** prepare the input sequence
	*/
    seq sequence( iprefix, ifill, isuffix );

    if ( sequence.getCount() == 0 ) {
	std::cerr << "No images in the input sequence: " << iprefix;
	std::cerr << std::endl;
	return __LINE__;
    }

	/* force a preload of the first image */
    (void) sequence.getBuffer( 0 );

    const PNG& src( sequence[ 0 ] );

	/* check width and height from the input image */
    if ( ww < 1 || hh < 1 ) {
	ww = src.getWidth();
	hh = src.getHeight();
    }

    unsigned int cntr = 0;
    double minx;
    double maxx;
    double miny;
    double maxy;
    double mint;
    double maxt;

    bool* hits = new bool[ ww * hh ];
    double* xoffs = new double[ ww * hh ];
    double* yoffs = new double[ ww * hh ];
    double* toffs = new double[ ww * hh ];

    scope ss( length );
    vec origin( 0.0, 0.0, 0.0, 0.0 );

    std::cerr << "Calculating offsets: " << std::flush;
    progress prepmeter( hh );

    for ( unsigned int jj=0; jj < hh; ++jj ) {
	double yy = - ( (double)jj / (double)(hh-1) ) + 0.5;

	prepmeter.tick();

	for ( unsigned int ii=0; ii < ww; ++ii ) {
	    double xx =  ( (double)ii / (double)(ww-1) ) - 0.5;

	    vec dir( 1.0, 2.0 * xx, 2.0 * yy, 1.0 );
	    ray in( origin, dir );
	    ray out;

	    unsigned int index = ii + jj * ww;
	    hits[ index ] = ss.reflect( in, &out );
	    if ( hits[ index ] ) {
		vec vv( out.getStart() );

		vv *= 1.0 / vv.getX();

		xoffs[ index ] = vv.getY();
		yoffs[ index ] = vv.getZ();
		toffs[ index ] = vv.getT();

		if ( ++cntr == 1 ) {
		    minx = maxx = xoffs[ index ];
		    miny = maxy = yoffs[ index ];
		    mint = maxt = toffs[ index ];
		} else {
		    if ( xoffs[ index ] < minx ) {
			minx = xoffs[ index ];
		    } else if ( maxx < xoffs[ index ] ) {
			maxx = xoffs[ index ];
		    }
		    if ( yoffs[ index ] < miny ) {
			miny = yoffs[ index ];
		    } else if ( maxy < yoffs[ index ] ) {
			maxy = yoffs[ index ];
		    }
		    if ( toffs[ index ] < mint ) {
			mint = toffs[ index ];
		    } else if ( maxt < toffs[ index ] ) {
			maxt = toffs[ index ];
		    }
		}
	    }
	}
    }

    std::cerr << " done" << std::endl;

    if ( cntr == 0 ) {
	std::cerr << "Something's wrong... we never hit the image.";
	std::cerr << std::endl;
	return __LINE__;
    }

    double xrange = ( maxx - minx );
    double yrange = ( maxy - miny );

    if ( std::fabs( xrange ) < 0.0001 || std::fabs( yrange ) < 0.0001 ) {
	std::cerr << "Something's wrong... we always hit center.";
	std::cerr << std::endl;
	return __LINE__;
    }

    double scale;

    if ( xrange < yrange ) {
	scale = 1.0 / yrange;
    } else {
	scale = 1.0 / xrange;
    }

    double tscale = fps / ( maxt - mint );

    std::cerr << "Scaling space offsets by " << scale << std::endl;
    std::cerr << "Scaling time offsets by " << tscale << std::endl;
    for ( unsigned int jj=0; jj < hh; ++jj ) {
	for ( unsigned int ii=0; ii < ww; ++ii ) {
	    unsigned int index = ii + jj * ww;
	    if ( hits[ index ] ) {
		xoffs[ index ] -= minx;
		xoffs[ index ] *= scale;

		yoffs[ index ] -= miny;
		yoffs[ index ] *= scale;

		toffs[ index ] -= mint;
		toffs[ index ] *= tscale / cc;
	    }
	}
    }

    maxx = ( maxx - minx ) * scale;
    minx = 0.0;

    maxy = ( maxy - miny ) * scale;
    miny = 0.0;

    maxt = ( maxt - mint ) * scale / cc;
    mint = 0.0;

	/*
	** prepare destination image
	*/
    PNG dst( ww, hh, 3 );

    unsigned int sw = src.getWidth();
    unsigned int sh = src.getHeight();
    unsigned int sc = src.getChannels();

	/* fake output sequence used to generate filenames */
    seq outseq( oprefix, ofill, osuffix );


    std::cerr << "Generating output: " << std::flush;
    progress outmeter( sequence.getCount() );

	/*
	** loop through to create each output frame
	*/
    for ( unsigned int ff=0; ff < sequence.getCount(); ++ff ) {
	unsigned char* dstPtr = dst.getBuffer();

	outmeter.tick();

	for ( unsigned int jj=0; jj < hh; ++jj ) {
	    for ( unsigned int ii=0; ii < ww; ++ii ) {
		unsigned int index = ii + jj * ww;
		if ( hits[ index ] ) {
		    unsigned int xx = clip( xoffs[ index ], (sw-1) );
		    unsigned int yy = clip( yoffs[ index ], (sh-1) );
		    unsigned int tt = (unsigned int)( toffs[ index ]+0.5 ) + ff;

		    unsigned char* srcPtr = sequence.getBuffer( tt );
		    if ( srcPtr != 0 ) {
			unsigned char* ptr = &srcPtr[ ( xx + yy * sw ) * sc ];

			*dstPtr++ = *ptr++;
			*dstPtr++ = *ptr++;
			*dstPtr++ = *ptr++;
		    } else {
			*dstPtr++ = 0x00;
			*dstPtr++ = 0x00;
			*dstPtr++ = 0x00;
		    }
		} else {
		    *dstPtr++ = 0xFF;
		    *dstPtr++ = 0xFF;
		    *dstPtr++ = 0xFF;
		}
	    }
	}

	    /* output the frame */
	char fname[ seq::MAXPATHLEN+1 ];
	outseq.filename( ff, fname, sizeof(fname) );
	if ( ! dst.write( fname ) ) {
	    std::cerr << "Unable to write: " << fname << std::endl;
	    std::exit( __LINE__ );
	}
    }

    std::cerr << " done" << std::endl;

    delete[] toffs;
    delete[] yoffs;
    delete[] xoffs;
    delete[] hits;

    return 0;
}
