
#include <cmath>

#include "vec.hh"
#include "ray.hh"
#include "mirror.hh"
#include "scope.hh"

scope::scope( double _length ) {
    static const double aa =  3.0 / 4.0;
    static const double bb = -1.0 / 4.0;

    this->add( aa, bb, bb, bb );
    this->add( bb, aa, bb, bb );
    this->add( bb, bb, aa, bb );
    this->add( bb, bb, bb, aa );

    this->length = _length;

    this->plane = mirror( vec( 1.0, 0.0, 0.0, 0.0 ), this->length );
}

scope::~scope( void ) {
    this->mirrors.clear();
}

bool scope::reflect( const ray& _in, ray* _out ) const {
    if ( this->reflect( _in, _out, this->mirrors.end() ) ) {
	ray rr( *_out );
	return this->plane.reflect( rr, _out );
    } else {
	return this->plane.reflect( _in, _out );
    }
}

bool scope::reflect( const ray& _in, ray* _out, iterator _skip ) const {
    iterator bestHit = this->mirrors.end();
    double bestDistance = -1.0;

    for ( iterator it=this->mirrors.begin(); it != this->mirrors.end(); ++it ) {
	if ( it != _skip ) {
	    double dd;
	    if ( it->checkRayIntersection( _in, &dd ) ) {
		if ( bestHit == this->mirrors.end() || dd < bestDistance ) {
		    bestHit = it;
		    bestDistance = dd;
		}
	    }
	}
    }

    if ( bestHit != this->mirrors.end() ) {
	ray reflected;
	bestHit->reflect( _in, &reflected );

	if ( reflected.getStart().getX() < this->length ) {
	    this->reflect( reflected, _out, bestHit );
	} else {
	    *_out = _in;
	}

	return true;
    } else {
	*_out = _in;
	return false;
    }
}

void scope::add( double _xx, double _yy, double _zz, double _tt ) {
    static const double alpha( 0.5 );
    static const double beta( M_SQRT1_2 );

    static const vec r1( alpha, alpha, alpha, alpha );
    static const vec r2( 0.0, 0.0, beta, -beta );
    static const vec r3( alpha, alpha, -alpha, -alpha );
    static const vec r4( beta, -beta, 0.0, 0.0 );

    static const double gamma( std::sqrt( 3.0 ) / 2.0 );

    vec vv( _xx, _yy, _zz, _tt );
    vec vp( r1.dot( vv ), r2.dot( vv ), r3.dot( vv ), r4.dot( vv ) );
    vp.normalize();

    mirror mm( vp, gamma );
    this->mirrors.push_front( mm );
}
